//
// BrackStack.go - автомат с магазинной памятью, распознающий вложенность
// круглых скобок.
// Построен на основе таблицы переходов созданной по следующей q-грамматике:
//
// 1. S -> (B)B
// 2. B -> (B)B
// 3. B -> empty
//

package main

import (
    "fmt"
    "os"
    "strings"
    "errors"
)

// Магазинными символами автомата являются S, B, маркер дна (botoom).
// Их значение должно быть за пределами видимых символов.
const (
  S = iota + 256
  B
  bottom
)

var str string      // Строка с входной цепочкой, имитирующая входную ленту
var i int           // Текущее положение входной головки
var brackets []rune // Срез, содержащий прочитанную строку со скобками

// Моделирование простого стека
type stack []int

func (s *stack) pop() (int, error) {
    theStack := *s
    if len(theStack) == 0 {
        return 0, errors.New("can't Pop() an empty stack")
    }
    x := theStack[len(theStack)-1]
    *s = theStack[:len(theStack)-1]
    return x, nil
}

func (s *stack) push(x int) {
    *s = append(*s, x)
}

func (s stack) top() (int, error) {
    if len(s) == 0 {
        return 0, errors.New("can't Top() an empty stack")
    }
    return s[len(s)-1], nil
}

func (s stack) empty() bool {
    return len(s) == 0
}

var stck stack  // Стек для магазинных символов

// Функция, реализующая чтение символов в буфер из входного потока.
// Используется для ввода с клавиатуры распознаваемой строки.
// Ввод осуществляется до нажатия на клавишу Enter.
// Символ '\n' яляется концевым маркером входной строки.
func GetOneLine() {
    fmt.Scanln(&str)
    str += "\n"; // Добавляется концевой маркер
}

// Инициализация устройств АМП
// Автоматически запускается до главной функции
func Init() {
    // Инициализация стека
    for(!stck.empty()) {
        stck.pop();
    }
    stck.push(bottom);
    stck.push(S);

    // Инициализация входной головки
    i = 0
}

// Устройство управления АМП, анализирующего вложенность скобок.
// Имитирует таблицу пееходов АМП.
func Parser() bool {
    // Инициализация устройств АМП
    Init()

    step := 0
    // Цикл анализа состояний
    for {
        x, err := stck.top()
        if err != nil {
            fmt.Println(err)
            os.Exit(2)
        }

        // Тестовый вывод информации о текущем шаге, текущем символе, вершине стека
        fmt.Print("step ", step, ": brackets[", i, "] = ", 
                  string(brackets[i]), " Top = ", x, "\n",
            )
        step++
        // Проверка содержимого на вершине стека
        switch x {
            // Анализ первой строки таблицы переходов
            case S:
                switch brackets[i] {
                    case '(': // [S, (]
                        stck.pop()
                        stck.push(B);
                        // эквивалентно 
                        //!!stck.top() = B  
                        stck.push(')')
                        stck.push(B)
                        i++             // следующий входной символ
                    case ')': // [S, )] - ) не может стоять в начале
                        fmt.Print("Position ", i, ", ", 
                                  "Error 1: ')' can not is in begin!\n",
                            )
                        return false
                    case '\n': // [S, концевой маркер]
                        fmt.Print("Position ", i, ", ",
                                  "Error 2: Empty string!\n",
                            )
                        return false
                    default: // Ошибка, такой символ не допустим
                        fmt.Print("Position ", i, ", ",
                                  "Incorrect symbol! ", string(brackets[i]), "\n",
                            )
                        return false
                }

            // Анализ второй строки таблицы переходов
            case B:
                switch(brackets[i]) {
                    case '(': // [B, (]
                        stck.pop()
                        stck.push(B)
                        // эквивалентно 
                        //!!stck.top() = B  
                        stck.push(')')
                        stck.push(B)
                        i++             // следующий входной символ
                    case ')': // [B, )] - вытолкнуть без сдвига 
                        stck.pop()
                    case '\n': // [B, концевой маркер] - вытолкнуть без сдвига
                        stck.pop()
                    default: // Ошибка, такой символ не допустим
                        fmt.Print("Position ", i, ", ",
                                  "Incorrect symbol! ", string(brackets[i]), "\n",
                            )
                        return false
                }

            case ')':
                // Анализ третьей строки таблицы переходов
                switch(brackets[i]) {
                    case '(': // [), (] - в принципе недостижимо
                        fmt.Print("Position ", i, ", ",
                                  "Error 3: I want to see')'!\n",
                            )
                        return false
                    case ')': // [), )] - вытолкнуть со сдвигом 
                        stck.pop() // вытолкнуть
                        i++        // сдвиг
                    case '\n': // [), концевой маркер]
                        fmt.Print("Position ", i, ", ",
                                  "Error 4: I want ')'!\n",
                            )
                        return false
                    default: // Ошибка, такой символ не допустим
                        fmt.Print("Position ", i, ", ",
                                  "Incorrect symbol! ", string(brackets[i]), "\n",
                            )
                        return false
                }

            // Анализ четвертой строки таблицы переходов
            case bottom:
                switch(str[i]) {
                    case '(': // [bottom, (] - невозможно при пустом магазине
                        fmt.Print("Position ", i, ", ",
                                  "Error 4: Impossible situation [bottom, '(']!\n",
                            )
                        return false
                    case ')': // [bottom, )] - лишняя )
                        fmt.Print("Position ", i, ", ",
                                  "Error 5: unnecessary ')'!\n",
                            )
                        return false
                    case '\n': // [bottom, концевой маркер] - допустить!
                        return true;
                    default: // Ошибка, такой символ не допустим
                        fmt.Print("Position ", i, ", ",
                                  "Incorrect symbol! ", string(brackets[i]), "\n",
                            )
                        return false
            }
        }
    }
}

//=============================================================================
// Главная функция используется для инициализации устройств АМП
// пред каждым новым прогоном и тестирования до тех пор, пока не 
// будет прочитана пустая строка
func main() {
    // Цикл распознавания различных входных цепочек
    for {
        strCursor := ""
        str = "";
        
        // Чтение очередной входной цепочки в буфер
        fmt.Print("Input bracket's expression: ")

        // Формируем очередную строку скобок для распознавания.
        GetOneLine()
        brackets = []rune(str)

        // Здесь начинается разбор принятой строки.
        if Parser() {
            fmt.Print("+++++ OK! +++++\n")
        } else {
            fmt.Print("----- Fatal error (look upper error message)! -----\n")
        }
        // Вывод разобранной строки и значения позиции входой головки.
        fmt.Print("Line: ", str)
        strCursor = " Pos: " + strings.Repeat("-", i)
        strCursor +="^"
        fmt.Print(strCursor, "  i = ", i, "\n\n")

        if str == "\n" {
            break
        }
    }
    fmt.Println("Goodbye!")
}
